@@ -5,17 +5,14 @@
 
 import json as _json
 import os as _os
-import random as _rd
 import shutil as _shutil
-import sys as _sys
 from os import sep as _sep
 from typing import List as _List
 
 import coinstac_dinunet.config as _conf
-import coinstac_dinunet.data.datautils as _du
 from coinstac_dinunet.config.keys import *
+from coinstac_dinunet.data import COINNDataHandle as _DataHandle
 from coinstac_dinunet.distrib import learner as _learner
-from coinstac_dinunet.profiler import Profile
 from coinstac_dinunet.utils import FrozenDict as _FrozenDict
 
 
@@ -35,11 +32,13 @@ class COINNLocal:
                  pin_memory: bool = False,
                  num_workers: int = 0,
                  load_limit: int = _conf.max_size,
+                 load_sparse=False,
                  pretrained_path: str = None,
                  patience: int = None,
                  num_folds: int = None,
                  split_ratio: _List[float] = None,
                  pretrain_args: dict = None,
+                 dataloader_args: dict = None,
                  **kw):
 
         self.out = {}
@@ -58,6 +57,7 @@ class COINNLocal:
         self._args['pin_memory'] = pin_memory
         self._args['num_workers'] = num_workers
         self._args['load_limit'] = load_limit
+        self._args['load_sparse'] = load_sparse
         self._args['pretrained_path'] = pretrained_path
         self._args['patience'] = patience if patience else epochs
         self._args['num_folds'] = num_folds
@@ -65,16 +65,16 @@ class COINNLocal:
         self._args.update(**kw)
         self._args = _FrozenDict(self._args)
         self._pretrain_args = pretrain_args if pretrain_args else {}
-        self._GLOBAL_STATE = {}
-        self.learner = None
+        self._dataloader_args = dataloader_args if dataloader_args else {}
 
     def _check_args(self):
         assert self.cache['computation_id'] is not None, self._PROMPT_TASK_
         assert self.cache['mode'] in [Mode.TRAIN, Mode.TEST], self._PROMPT_MODE_
 
-    def _init_runs(self):
+    def _init_runs(self, learner):
         out = {}
-        out.update(_du.init_k_folds(self.cache, self.state))
+        """Data related initializations."""
+        out.update(learner.trainer.data_handle.prepare_data())
         out['data_size'] = {}
         for k, sp in self.cache['splits'].items():
             sp = _json.loads(open(self.cache['split_dir'] + _os.sep + sp).read())
@@ -83,51 +83,65 @@ class COINNLocal:
             out[k] = self.cache.get(k)
         return out
 
-    def _init_nn_state(self, trainer):
+    def _attach_global(self, learner):
+        self.cache['nn'] = learner.trainer.nn
+        self.cache['device'] = learner.trainer.device
+        self.cache['optimizer'] = learner.trainer.optimizer
+        self.cache['dataset'] = learner.trainer.data_handle.dataset
+
+    def _next_run(self, learner):
         out = {}
-        self.cache['current_nn_state'] = 'current.distrib.pt'
-        self.cache['best_nn_state'] = 'best.distrib.pt'
-        trainer.init_nn(init_weights=True)
-        trainer.save_checkpoint(file_path=self.cache['log_dir'] + _sep + self.cache['current_nn_state'])
+        learner.trainer.init_nn(init_weights=True)
+        self.cache['best_nn_state'] = f"best.{self.cache['computation_id']}-{self.cache['split_ix']}.pt"
         out['phase'] = Phase.COMPUTATION
+        self._attach_global(learner)
         return out
 
-    def _pretrain_local(self, trainer_cls, dataset_cls):
-        out = {}
+    def _pretrain_local(self, trainer_cls, train_dataset, validation_dataset):
+        out = {'phase': Phase.COMPUTATION}
         if self._pretrain_args.get('epochs') and self.cache['pretrain']:
             cache = {**self.cache}
             cache.update(**self._pretrain_args)
             trainer = trainer_cls(cache=cache, input=self.input, state=self.state)
             trainer.init_nn()
+
             trainer.init_training_cache()
-            out.update(**trainer.train_local(dataset_cls))
-            _rd.shuffle(trainer.cache.get('data_indices', []))
+            out.update(**trainer.train_local(train_dataset, validation_dataset))
             out['phase'] = Phase.PRE_COMPUTATION
 
-        if self._pretrain_args.get('epochs') and any([r['pretrain'] for r in self._GLOBAL_STATE['runs'].values()]):
+        if self._pretrain_args.get('epochs') and any([r['pretrain'] for r in self.input['global_runs'].values()]):
             out['phase'] = Phase.PRE_COMPUTATION
         return out
 
-    @Profile()
-    def compute(self, dataset_cls, trainer_cls, learner_cls: callable = None, **kw):
-        self.out['phase'] = self.input.get('phase', Phase.INIT_RUNS)
-        trainer = trainer_cls(cache=self.cache, input=self.input, state=self.state)
+    def compute(self, trainer_cls,
+                dataset_cls=None,
+                datahandle_cls=_DataHandle,
+                learner_cls=_learner.COINNLearner,
+                **kw):
 
+        learner = learner_cls(
+            trainer=trainer_cls(
+                data_handle=datahandle_cls(
+                    cache=self.cache, input=self.input, state=self.state,
+                    dataloader_args=self._dataloader_args
+                )
+            )
+        )
+
+        self.out['phase'] = self.input.get('phase', Phase.INIT_RUNS)
         if self.out['phase'] == Phase.INIT_RUNS:
             """ Generate folds as specified.   """
             self.cache.update(**self.input)
             for k in self._args:
                 if self.cache.get(k) is None:
                     self.cache[k] = self._args[k]
-            self.out.update(**self._init_runs())
+            self.out.update(**self._init_runs(learner))
             self.cache['args'] = _FrozenDict({**self.cache})
             self.cache['verbose'] = False
             self._check_args()
 
-        elif self.out['phase'] == Phase.INIT_NN:
-            """  Initialize neural network/optimizer and GPUs  """
-            self._GLOBAL_STATE['runs'] = self.input['global_runs']
-            self.cache.update(**self._GLOBAL_STATE['runs'][self.state['clientId']])
+        elif self.out['phase'] == Phase.NEXT_RUN:
+            self.cache.update(**self.input['global_runs'][self.state['clientId']])
             self.cache.update(cursor=0)
             self.cache[Key.TRAIN_SERIALIZABLE] = []
 
@@ -135,58 +149,51 @@ class COINNLocal:
             self.cache['log_dir'] = self.state['outputDirectory'] + _sep + self.cache[
                 'computation_id'] + _sep + f"fold_{self.cache['split_ix']}"
             _os.makedirs(self.cache['log_dir'], exist_ok=True)
-            self.out.update(**self._init_nn_state(trainer))
-            trainer.cache_data_indices(dataset_cls, split_key='train')
-            self.out.update(**self._pretrain_local(trainer_cls, dataset_cls))
+
+            self.out.update(**self._next_run(learner))
+            self.out.update(
+                **self._pretrain_local(
+                    trainer_cls,
+                    learner.trainer.data_handle.get_train_dataset(dataset_cls),
+                    learner.trainer.data_handle.get_validation_dataset(dataset_cls))
+            )
 
         elif self.out['phase'] == Phase.PRE_COMPUTATION and self.input.get('pretrained_weights'):
-            trainer.init_nn(init_weights=False)
-            trainer.load_checkpoint(
-                file_path=self.state['baseDirectory'] + _sep + self.input['pretrained_weights'])
-            trainer.save_checkpoint(
-                file_path=self.cache['log_dir'] + _sep + self.cache['current_nn_state'])
+            learner.trainer.load_checkpoint(
+                file_path=self.state['baseDirectory'] + _sep + self.input['pretrained_weights']
+            )
             self.out['phase'] = Phase.COMPUTATION
 
-        """################################### Computation ##########################################"""
-        self._GLOBAL_STATE['modes'] = self.input.get('global_modes', {})
-        self.out['mode'] = self._GLOBAL_STATE['modes'].get(self.state['clientId'], self.cache['mode'])
+        """Track global state among sites."""
+        self.out['mode'] = learner.global_modes.get(self.state['clientId'], self.cache['mode'])
 
+        """Computation begins..."""
         if self.out['phase'] == Phase.COMPUTATION:
+
             """ Train/validation and test phases """
-            trainer.init_nn(init_weights=False)
-
-            trainer.load_checkpoint(
-                file_path=self.cache['log_dir'] + _sep + self.cache['current_nn_state'])
-
             if self.input.get('save_current_as_best'):
-                trainer.save_checkpoint(
-                    file_path=self.cache['log_dir'] + _sep + self.cache['best_nn_state'])
+                learner.trainer.save_checkpoint(
+                    file_path=self.cache['log_dir'] + _sep + self.cache['best_nn_state']
+                )
 
             """Initialize Learner and assign trainer"""
-            self._set_learner(learner_cls, trainer=trainer, **kw)
-
-            """ Reducer must issue update signal for the network to update"""
             if self.input.get('update'):
-                self.out.update(**self.learner.step())
-                if self.out['save_state']:
-                    trainer.save_checkpoint(
-                        file_path=self.cache['log_dir'] + _sep + self.cache['current_nn_state']
-                    )
+                self.out.update(**learner.step())
 
-            if any(m == Mode.TRAIN for m in self._GLOBAL_STATE['modes'].values()):
+            if any(m == Mode.TRAIN for m in learner.global_modes.values()):
                 """
                 All sites must begin/resume the training the same time.
                 To enforce this, we have a 'val_waiting' status. Lagged sites will go to this status, 
                    and reshuffle the data,
                 take part in the training with everybody until all sites go to 'val_waiting' status.
                 """
-                out, it = self.learner.to_reduce(dataset_cls)
+                it, out = learner.to_reduce()
                 self.out.update(**out)
                 if it.get('averages') and it.get('metrics'):
                     self.cache[Key.TRAIN_SERIALIZABLE].append([vars(it['averages']), vars(it['metrics'])])
-                    self.out.update(**trainer.on_iteration_end(0, 0, it))
+                    self.out.update(**learner.trainer.on_iteration_end(0, 0, it))
 
-            if all(m == Mode.VALIDATION for m in self._GLOBAL_STATE['modes'].values()):
+            if all(m == Mode.VALIDATION for m in learner.global_modes.values()):
                 """
                 Once all sites are in 'val_waiting' status, remote issues 'validation' signal.
                 Once all sites run validation phase, they go to 'train_waiting' status.
@@ -194,30 +201,20 @@ class COINNLocal:
                  and all sites reshuffle the indices and r esume training.
                 We send the confusion matrix to the remote to accumulate global score for model selection.
                 """
-                self.out.update(**trainer.validation_distributed(dataset_cls))
+                self.out.update(**learner.trainer.validation_distributed(dataset_cls))
                 self.out['mode'] = Mode.TRAIN_WAITING
 
-            if all(m == Mode.TEST for m in self._GLOBAL_STATE['modes'].values()):
-                self.out.update(**trainer.test_distributed(dataset_cls))
+            if all(m == Mode.TEST for m in learner.global_modes.values()):
+                self.out.update(**learner.trainer.test_distributed(dataset_cls))
                 self.out['mode'] = self.cache['args']['mode']
                 self.out['phase'] = Phase.NEXT_RUN_WAITING
 
         elif self.out['phase'] == Phase.SUCCESS:
             """ This phase receives global scores from the aggregator."""
             _shutil.copy(f"{self.state['baseDirectory']}{_sep}{self.input['results_zip']}.zip",
-                         f"{self.state['outputDirectory'] + _sep + self.cache['computation_id']}{_sep}{self.input['results_zip']}.zip")
+                         f"{self.state['outputDirectory'] + _sep + self.cache['computation_id']}{_sep}"
+                         f"{self.input['results_zip']}.zip")
 
-    def _set_learner(self, learner_cls: _learner.COINNLearner = None, trainer=None, **kw):
-
-        if learner_cls is None:
-            learner_cls = _learner.COINNLearner
-
-        self.learner = learner_cls(trainer=trainer, global_state=self._GLOBAL_STATE, **kw)
-
-    def send(self):
-        output = {'output': self.out, 'cache': self.cache}
-        try:
-            output = _json.dumps(output)
-            _sys.stdout.write(output)
-        except Exception as e:
-            raise Exception(f"Error parsing Json at {self.state['clientId']} {e}:\n", output)
+    def __call__(self, *args, **kwargs):
+        self.compute(*args, **kwargs)
+        return {'output': self.out}
